// Copyright (C) 2019  Wilko Manger
//
// This file is part of url.
//
// url is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// url is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with url.  If not, see <https://www.gnu.org/licenses/>.
library url;

import 'package:validators/validators.dart';

class Url implements Uri {

  final Uri _uri;

  static const _protocols = [
    'http',
    'https',
    'ftp',
    'mxc'
  ];

  /// Create an [Url] from an [Uri].
  Url(this._uri) {
    if (!isUrl(_uri)) {
      throw ArgumentError('Uri is not an URL');
    }
  }

  Url._noCheck(this._uri);

  /// Parse [Uri], requiring that is an URL, not just an [Uri].
  ///
  /// Will accept `mxc` (Matrix Content) URLs.
  ///
  /// Note that the returned type is still an [Uri], but guaranteed
  /// that it's an URL.
  factory Url.parse(String url) {
    if (!isUrl(url)) {
      throw FormatException('Not a valid url: $url');
    }

    return Url._noCheck(Uri.parse(url));
  }

  @override
  String get authority => _uri.authority;

  @override
  UriData get data => _uri.data;

  @override
  String get fragment => _uri.fragment;

  @override
  bool get hasAbsolutePath => _uri.hasAbsolutePath;

  @override
  bool get hasAuthority => _uri.hasAuthority;

  @override
  bool get hasEmptyPath => _uri.hasEmptyPath;

  @override
  bool get hasFragment => _uri.hasFragment;

  @override
  bool get hasPort => _uri.hasPort;

  @override
  bool get hasQuery => _uri.hasQuery;

  @override
  bool get hasScheme => _uri.hasScheme;

  @override
  String get host => _uri.host;

  @override
  bool get isAbsolute => _uri.isAbsolute;

  @override
  bool isScheme(String scheme) => _uri.isScheme(scheme);

  @override
  Url normalizePath() => Url._noCheck(_uri.normalizePath());

  @override
  String get origin => _uri.origin;

  @override
  String get path => _uri.path;

  @override
  List<String> get pathSegments => _uri.pathSegments;

  @override
  int get port => _uri.port;

  @override
  String get query => _uri.query;

  @override
  Map<String, String> get queryParameters => _uri.queryParameters;

  @override
  Map<String, List<String>> get queryParametersAll => _uri.queryParametersAll;

  @override
  Url removeFragment() => Url._noCheck(_uri.removeFragment());

  @override
  Url replace({
    String scheme,
    String userInfo,
    String host,
    int port,
    String path,
    Iterable<String> pathSegments,
    String query,
    Map<String, dynamic> queryParameters,
    String fragment}) => Url._noCheck(_uri.replace(
      scheme: scheme,
      userInfo: userInfo,
      host: host,
      port: port,
      path: path,
      pathSegments: pathSegments,
      query: query,
      queryParameters: queryParameters,
      fragment: fragment
    ));

  @override
  Url resolve(String reference) => Url._noCheck(_uri.resolve(reference));
  
  @override
  Url resolveUri(Uri reference) => Url._noCheck(_uri.resolveUri(reference));

  /// Resolve [reference] as an URL relative to `this`.
  ///
  /// Returns the resolved URL.
  Url resolveUrl(Url reference) => resolveUri(reference._uri);
  
  @override
  String get scheme => _uri.scheme;
  
  @override
  String toFilePath({bool windows}) => _uri.toFilePath(windows: windows);
  
  @override
  String get userInfo => _uri.userInfo;

  @override
  String toString() => _uri.toString();

  @override
  bool operator ==(Object other) => _uri == other;

  @override
  int get hashCode => _uri.hashCode;
}

/// True if [input] is an URL, false otherwise.
///
/// [input] can be an [Uri] or [String].
bool isUrl(dynamic input) {
  var str;
  if (input is Uri) {
    str = input.toString();
  } else if (input is String) {
    str = input;
  }

  // input was neither Uri nor String
  if (str == null) {
    return false;
  } else {
    return isURL(
      input,
      protocols: Url._protocols,
      requireProtocol: true
    );
  }
}
